@@ -41,66 +41,11 @@
 #include <iostream>
 #endif
 
-#ifndef COMPUTE_Z
-#define COMPUTE_Z 1
-#endif // COMPUTE_Z
-
-using namespace std;
-
 using namespace geos::geom;
 
 namespace geos {
 namespace algorithm { // geos.algorithm
 
-namespace { // anonymous
-
-/**
- * Finds the endpoint of the segments P and Q which
- * is closest to the other segment.
- * This is a reasonable surrogate for the true
- * intersection points in ill-conditioned cases
- * (e.g. where two segments are nearly coincident,
- * or where the endpoint of one segment lies almost on the other segment).
- * <p>
- * This replaces the older CentralEndpoint heuristic,
- * which chose the wrong endpoint in some cases
- * where the segments had very distinct slopes
- * and one endpoint lay almost on the other segment.
- *
- * @param p1 an endpoint of segment P
- * @param p2 an endpoint of segment P
- * @param q1 an endpoint of segment Q
- * @param q2 an endpoint of segment Q
- * @return the nearest endpoint to the other segment
- */
-Coordinate
-nearestEndpoint(const Coordinate& p1, const Coordinate& p2,
-                const Coordinate& q1, const Coordinate& q2)
-{
-    const Coordinate* nearestPt = &p1;
-    double minDist = Distance::pointToSegment(p1, q1, q2);
-
-    double dist = Distance::pointToSegment(p2, q1, q2);
-    if(dist < minDist) {
-        minDist = dist;
-        nearestPt = &p2;
-    }
-    dist = Distance::pointToSegment(q1, p1, p2);
-    if(dist < minDist) {
-        minDist = dist;
-        nearestPt = &q1;
-    }
-    dist = Distance::pointToSegment(q2, p1, p2);
-    if(dist < minDist) {
-        minDist = dist;
-        nearestPt = &q2;
-    }
-    return *nearestPt;
-}
-
-
-} // anonymous namespace
-
 /*public static*/
 double
 LineIntersector::computeEdgeDistance(const Coordinate& p, const Coordinate& p0, const Coordinate& p1)
@@ -152,10 +97,10 @@ LineIntersector::computeIntersection(const Coordinate& p1, const Coordinate& p2,
 }
 
 /*public*/
-string
+std::string
 LineIntersector::toString() const
 {
-    string str = inputLines[0][0]->toString() + "_"
+    std::string str = inputLines[0][0]->toString() + "_"
                  + inputLines[0][1]->toString() + " "
                  + inputLines[1][0]->toString() + "_"
                  + inputLines[1][1]->toString() + " : ";
@@ -189,21 +134,10 @@ LineIntersector::computeIntLineIndex()
     computeIntLineIndex(1);
 }
 
-/*public*/
-bool
-LineIntersector::isIntersection(const Coordinate& pt) const
-{
-    for(size_t i = 0; i < result; ++i) {
-        if(intPt[i].equals2D(pt)) {
-            return true;
-        }
-    }
-    return false;
-}
 
 /*public*/
 const Coordinate&
-LineIntersector::getIntersectionAlongSegment(int segmentIndex, int intIndex)
+LineIntersector::getIntersectionAlongSegment(std::size_t segmentIndex, std::size_t intIndex)
 {
     // lazily compute int line array
     computeIntLineIndex();
@@ -211,8 +145,8 @@ LineIntersector::getIntersectionAlongSegment(int segmentIndex, int intIndex)
 }
 
 /*public*/
-int
-LineIntersector::getIndexAlongSegment(int segmentIndex, int intIndex)
+size_t
+LineIntersector::getIndexAlongSegment(std::size_t segmentIndex, std::size_t intIndex)
 {
     computeIntLineIndex();
     return intLineIndex[segmentIndex][intIndex];
@@ -220,7 +154,7 @@ LineIntersector::getIndexAlongSegment(int segmentIndex, int intIndex)
 
 /*private*/
 void
-LineIntersector::computeIntLineIndex(int segmentIndex)
+LineIntersector::computeIntLineIndex(std::size_t segmentIndex)
 {
     double dist0 = getEdgeDistance(segmentIndex, 0);
     double dist1 = getEdgeDistance(segmentIndex, 1);
@@ -236,7 +170,7 @@ LineIntersector::computeIntLineIndex(int segmentIndex)
 
 /*public*/
 double
-LineIntersector::getEdgeDistance(size_t segmentIndex, size_t intIndex) const
+LineIntersector::getEdgeDistance(std::size_t segmentIndex, std::size_t intIndex) const
 {
     double dist = computeEdgeDistance(intPt[intIndex],
                                       *inputLines[segmentIndex][0],
@@ -244,74 +178,48 @@ LineIntersector::getEdgeDistance(size_t segmentIndex, size_t intIndex) const
     return dist;
 }
 
-/*public*/
-bool
-LineIntersector::isInteriorIntersection()
-{
-    if(isInteriorIntersection(0)) {
-        return true;
-    }
-    if(isInteriorIntersection(1)) {
-        return true;
-    }
-    return false;
-}
-
-/*public*/
-bool
-LineIntersector::isInteriorIntersection(int inputLineIndex)
-{
-    for(size_t i = 0; i < result; ++i) {
-        if(!(intPt[i].equals2D(*inputLines[inputLineIndex][0])
-                || intPt[i].equals2D(*inputLines[inputLineIndex][1]))) {
-            return true;
-        }
-    }
-    return false;
-}
-
 /*public static*/
 double
 LineIntersector::interpolateZ(const Coordinate& p,
                               const Coordinate& p1, const Coordinate& p2)
 {
 #if GEOS_DEBUG
-    cerr << "LineIntersector::interpolateZ(" << p.toString() << ", " << p1.toString() << ", " << p2.toString() << ")" <<
-         endl;
+    std::cerr << "LineIntersector::interpolateZ(" << p.toString() << ", " << p1.toString() << ", " << p2.toString() << ")" <<
+         std::endl;
 #endif
 
     if(std::isnan(p1.z)) {
 #if GEOS_DEBUG
-        cerr << " p1 do not have a Z" << endl;
+        std::cerr << " p1 do not have a Z" << std::endl;
 #endif
         return p2.z; // might be DoubleNotANumber again
     }
 
     if(std::isnan(p2.z)) {
 #if GEOS_DEBUG
-        cerr << " p2 do not have a Z" << endl;
+        std::cerr << " p2 do not have a Z" << std::endl;
 #endif
         return p1.z; // might be DoubleNotANumber again
     }
 
     if(p == p1) {
 #if GEOS_DEBUG
-        cerr << " p==p1, returning " << p1.z << endl;
+        std::cerr << " p==p1, returning " << p1.z << std::endl;
 #endif
         return p1.z;
     }
     if(p == p2) {
 #if GEOS_DEBUG
-        cerr << " p==p2, returning " << p2.z << endl;
+        std::cerr << " p==p2, returning " << p2.z << std::endl;
 #endif
         return p2.z;
     }
 
     //double zgap = fabs(p2.z - p1.z);
     double zgap = p2.z - p1.z;
-    if(! zgap) {
+    if(zgap == 0.0) {
 #if GEOS_DEBUG
-        cerr << " no zgap, returning " << p2.z << endl;
+        std::cerr << " no zgap, returning " << p2.z << std::endl;
 #endif
         return p2.z;
     }
@@ -326,8 +234,8 @@ LineIntersector::interpolateZ(const Coordinate& p,
     //double interpolated = p1.z < p2.z ? p1.z+zoff : p1.z-zoff;
     double interpolated = p1.z + zoff;
 #if GEOS_DEBUG
-    cerr << " zgap:" << zgap << " seglen:" << seglen << " pdist:" << pdist
-         << " fract:" << fract << " z:" << interpolated << endl;
+    std::cerr << " zgap:" << zgap << " seglen:" << seglen << " pdist:" << pdist
+         << " fract:" << fract << " z:" << interpolated << std::endl;
 #endif
     return interpolated;
 
@@ -348,21 +256,6 @@ LineIntersector::computeIntersection(const Coordinate& p, const Coordinate& p1,
             if((p == p1) || (p == p2)) { // 2d only test
                 isProperVar = false;
             }
-#if COMPUTE_Z
-            intPt[0] = p;
-#if GEOS_DEBUG
-            cerr << "RobustIntersector::computeIntersection(Coordinate,Coordinate,Coordinate) calling interpolateZ" << endl;
-#endif
-            double z = interpolateZ(p, p1, p2);
-            if(!std::isnan(z)) {
-                if(std::isnan(intPt[0].z)) {
-                    intPt[0].z = z;
-                }
-                else {
-                    intPt[0].z = (intPt[0].z + z) / 2;
-                }
-            }
-#endif // COMPUTE_Z
             result = POINT_INTERSECTION;
             return;
         }
@@ -384,14 +277,14 @@ LineIntersector::hasIntersection(const Coordinate& p, const Coordinate& p1, cons
 }
 
 /*private*/
-int
+uint8_t
 LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
                                   const Coordinate& q1, const Coordinate& q2)
 {
 #if GEOS_DEBUG
-    cerr << "LineIntersector::computeIntersect called" << endl;
-    cerr << " p1:" << p1.toString() << " p2:" << p2.toString() << " q1:" << q1.toString() << " q2:" << q2.toString() <<
-         endl;
+    std::cerr << "LineIntersector::computeIntersect called" << std::endl;
+    std::cerr << " p1:" << p1.toString() << " p2:" << p2.toString() << " q1:" << q1.toString() << " q2:" << q2.toString() <<
+         std::endl;
 #endif // GEOS_DEBUG
 
     isProperVar = false;
@@ -399,7 +292,7 @@ LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
     // first try a fast test to see if the envelopes of the lines intersect
     if(!Envelope::intersects(p1, p2, q1, q2)) {
 #if GEOS_DEBUG
-        cerr << " NO_INTERSECTION" << endl;
+        std::cerr << " NO_INTERSECTION" << std::endl;
 #endif
         return NO_INTERSECTION;
     }
@@ -412,7 +305,7 @@ LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
 
     if((Pq1 > 0 && Pq2 > 0) || (Pq1 < 0 && Pq2 < 0)) {
 #if GEOS_DEBUG
-        cerr << " NO_INTERSECTION" << endl;
+        std::cerr << " NO_INTERSECTION" << std::endl;
 #endif
         return NO_INTERSECTION;
     }
@@ -422,20 +315,23 @@ LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
 
     if((Qp1 > 0 && Qp2 > 0) || (Qp1 < 0 && Qp2 < 0)) {
 #if GEOS_DEBUG
-        cerr << " NO_INTERSECTION" << endl;
+        std::cerr << " NO_INTERSECTION" << std::endl;
 #endif
         return NO_INTERSECTION;
     }
 
+    /**
+     * Intersection is collinear if each endpoint lies on the other line.
+     */
     bool collinear = Pq1 == 0 && Pq2 == 0 && Qp1 == 0 && Qp2 == 0;
     if(collinear) {
 #if GEOS_DEBUG
-        cerr << " computingCollinearIntersection" << endl;
+        std::cerr << " computingCollinearIntersection" << std::endl;
 #endif
         return computeCollinearIntersection(p1, p2, q1, q2);
     }
 
-    /**
+    /*
      * At this point we know that there is a single intersection point
      * (since the lines are not collinear).
      */
@@ -450,13 +346,17 @@ LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
      * since at this point we know that the inputLines must
      *  intersect.
      */
+    Coordinate p;
+    double z = DoubleNotANumber;
+
     if(Pq1 == 0 || Pq2 == 0 || Qp1 == 0 || Qp2 == 0) {
-#if COMPUTE_Z
-        int hits = 0;
-        double z = 0.0;
-#endif
+
         isProperVar = false;
 
+#if GEOS_DEBUG
+        std::cerr << " intersection is NOT proper" << std::endl;
+#endif
+
         /* Check for two equal endpoints.
          * This is done explicitly rather than by the orientation tests
          * below in order to improve robustness.
@@ -470,371 +370,157 @@ LineIntersector::computeIntersect(const Coordinate& p1, const Coordinate& p2,
          * LINESTRING ( -48.51001596420236 -22.063180333403878,
          * 			19.850257749638203 46.29709338043669 )
          *
-         * which used to produce the result:
+         * which used to produce the INCORRECT result:
          * (20.31970698357233, 46.76654261437082, NaN)
          */
 
-        if(p1.equals2D(q1) || p1.equals2D(q2)) {
-            intPt[0] = p1;
-#if COMPUTE_Z
-            if(!std::isnan(p1.z)) {
-                z += p1.z;
-                hits++;
-            }
-#endif
+        if (p1.equals2D(q1)) {
+            p = p1;
+            z = zGet(p1, q1);
         }
-        else if(p2.equals2D(q1) || p2.equals2D(q2)) {
-            intPt[0] = p2;
-#if COMPUTE_Z
-            if(!std::isnan(p2.z)) {
-                z += p2.z;
-                hits++;
-            }
-#endif
+        else if (p1.equals2D(q2)) {
+            p = p1;
+            z = zGet(p1, q2);
         }
-
-        /**
+        else if (p2.equals2D(q1)) {
+            p = p2;
+            z = zGet(p2, q1);
+        }
+        else if (p2.equals2D(q2)) {
+            p = p2;
+            z = zGet(p2, q2);
+        }
+        /*
          * Now check to see if any endpoint lies on the interior of the other segment.
          */
         else if(Pq1 == 0) {
-            intPt[0] = q1;
-#if COMPUTE_Z
-            if(!std::isnan(q1.z)) {
-                z += q1.z;
-                hits++;
-            }
-#endif
+            p = q1;
+            z = zGetOrInterpolate(q1, p1, p2);
         }
         else if(Pq2 == 0) {
-            intPt[0] = q2;
-#if COMPUTE_Z
-            if(!std::isnan(q2.z)) {
-                z += q2.z;
-                hits++;
-            }
-#endif
+            p = q2;
+            z = zGetOrInterpolate(q2, p1, p2);
         }
         else if(Qp1 == 0) {
-            intPt[0] = p1;
-#if COMPUTE_Z
-            if(!std::isnan(p1.z)) {
-                z += p1.z;
-                hits++;
-            }
-#endif
+            p = p1;
+            z = zGetOrInterpolate(p1, q1, q2);
         }
         else if(Qp2 == 0) {
-            intPt[0] = p2;
-#if COMPUTE_Z
-            if(!std::isnan(p2.z)) {
-                z += p2.z;
-                hits++;
-            }
-#endif
+            p = p2;
+            z = zGetOrInterpolate(p2, q1, q2);
         }
-#if COMPUTE_Z
-#if GEOS_DEBUG
-        cerr << "LineIntersector::computeIntersect: z:" << z << " hits:" << hits << endl;
-#endif // GEOS_DEBUG
-        if(hits) {
-            intPt[0].z = z / hits;
-        }
-#endif // COMPUTE_Z
     }
     else {
-        isProperVar = true;
-        intPt[0] = intersection(p1, p2, q1, q2);
-    }
 #if GEOS_DEBUG
-    cerr << " POINT_INTERSECTION; intPt[0]:" << intPt[0].toString() << endl;
+        std::cerr << " intersection is proper" << std::endl;
+#endif // GEOS_DEBUG
+        isProperVar = true;
+        p = intersection(p1, p2, q1, q2);
+#if GEOS_DEBUG
+        std::cerr << " computed intersection point: " << p << std::endl;
+#endif // GEOS_DEBUG
+        z = zInterpolate(p, p1, p2, q1, q2);
+#if GEOS_DEBUG
+        std::cerr << " computed proper intersection Z: " << z << std::endl;
+#endif // GEOS_DEBUG
+    }
+    intPt[0] = Coordinate(p.x, p.y, z);
+#if GEOS_DEBUG
+    std::cerr << " POINT_INTERSECTION; intPt[0]:" << intPt[0].toString() << std::endl;
 #endif // GEOS_DEBUG
     return POINT_INTERSECTION;
 }
 
 /*private*/
-int
+uint8_t
 LineIntersector::computeCollinearIntersection(const Coordinate& p1, const Coordinate& p2,
                                               const Coordinate& q1, const Coordinate& q2)
 {
-#if COMPUTE_Z
-    double ztot;
-    int hits;
-    double p2z;
-    double p1z;
-    double q1z;
-    double q2z;
-#endif // COMPUTE_Z
 
 #if GEOS_DEBUG
-    cerr << "LineIntersector::computeCollinearIntersection called" << endl;
-    cerr << " p1:" << p1.toString() << " p2:" << p2.toString() << " q1:" << q1.toString() << " q2:" << q2.toString() <<
-         endl;
+    std::cerr << "LineIntersector::computeCollinearIntersection called" << std::endl;
+    std::cerr << " p1:" << p1.toString() << " p2:" << p2.toString() << " q1:" << q1.toString() << " q2:" << q2.toString() <<
+         std::endl;
 #endif // GEOS_DEBUG
 
-    bool p1q1p2 = Envelope::intersects(p1, p2, q1);
-    bool p1q2p2 = Envelope::intersects(p1, p2, q2);
-    bool q1p1q2 = Envelope::intersects(q1, q2, p1);
-    bool q1p2q2 = Envelope::intersects(q1, q2, p2);
+    bool q1inP = Envelope::intersects(p1, p2, q1);
+    bool q2inP = Envelope::intersects(p1, p2, q2);
+    bool p1inQ = Envelope::intersects(q1, q2, p1);
+    bool p2inQ = Envelope::intersects(q1, q2, p2);
 
-    if(p1q1p2 && p1q2p2) {
+    if(q1inP && q2inP) {
 #if GEOS_DEBUG
-        cerr << " p1q1p2 && p1q2p2" << endl;
-#endif
-        intPt[0] = q1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q1z = interpolateZ(q1, p1, p2);
-        if(!std::isnan(q1z)) {
-            ztot += q1z;
-            hits++;
-        }
-        if(!std::isnan(q1.z)) {
-            ztot += q1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = q2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q2z = interpolateZ(q2, p1, p2);
-        if(!std::isnan(q2z)) {
-            ztot += q2z;
-            hits++;
-        }
-        if(!std::isnan(q2.z)) {
-            ztot += q2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " q1inP && q2inP" << std::endl;
 #endif
+        intPt[0] = zGetOrInterpolateCopy(q1, p1, p2);
+        intPt[1] = zGetOrInterpolateCopy(q2, p1, p2);
 #if GEOS_DEBUG
-        cerr << " intPt[0]: " << intPt[0].toString() << endl;
-        cerr << " intPt[1]: " << intPt[1].toString() << endl;
+        std::cerr << " intPt[0]: " << intPt[0].toString() << std::endl;
+        std::cerr << " intPt[1]: " << intPt[1].toString() << std::endl;
 #endif
         return COLLINEAR_INTERSECTION;
     }
-    if(q1p1q2 && q1p2q2) {
+    if(p1inQ && p2inQ) {
 #if GEOS_DEBUG
-        cerr << " q1p1q2 && q1p2q2" << endl;
-#endif
-        intPt[0] = p1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p1z = interpolateZ(p1, q1, q2);
-        if(!std::isnan(p1z)) {
-            ztot += p1z;
-            hits++;
-        }
-        if(!std::isnan(p1.z)) {
-            ztot += p1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = p2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p2z = interpolateZ(p2, q1, q2);
-        if(!std::isnan(p2z)) {
-            ztot += p2z;
-            hits++;
-        }
-        if(!std::isnan(p2.z)) {
-            ztot += p2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " p1inQ && p2inQ" << std::endl;
 #endif
+        intPt[0] = zGetOrInterpolateCopy(p1, q1, q2);
+        intPt[1] = zGetOrInterpolateCopy(p2, q1, q2);
         return COLLINEAR_INTERSECTION;
     }
-    if(p1q1p2 && q1p1q2) {
+    if(q1inP && p1inQ) {
 #if GEOS_DEBUG
-        cerr << " p1q1p2 && q1p1q2" << endl;
-#endif
-        intPt[0] = q1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q1z = interpolateZ(q1, p1, p2);
-        if(!std::isnan(q1z)) {
-            ztot += q1z;
-            hits++;
-        }
-        if(!std::isnan(q1.z)) {
-            ztot += q1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = p1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p1z = interpolateZ(p1, q1, q2);
-        if(!std::isnan(p1z)) {
-            ztot += p1z;
-            hits++;
-        }
-        if(!std::isnan(p1.z)) {
-            ztot += p1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " q1inP && p1inQ" << std::endl;
 #endif
+        // if pts are equal Z is chosen arbitrarily
+        intPt[0] = zGetOrInterpolateCopy(q1, p1, p2);
+        intPt[1] = zGetOrInterpolateCopy(p1, q1, q2);
+
 #if GEOS_DEBUG
-        cerr << " intPt[0]: " << intPt[0].toString() << endl;
-        cerr << " intPt[1]: " << intPt[1].toString() << endl;
+        std::cerr << " intPt[0]: " << intPt[0].toString() << std::endl;
+        std::cerr << " intPt[1]: " << intPt[1].toString() << std::endl;
 #endif
-        return (q1 == p1) && !p1q2p2 && !q1p2q2 ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
+        return (q1 == p1) && !q2inP && !p2inQ ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
     }
-    if(p1q1p2 && q1p2q2) {
+    if(q1inP && p2inQ) {
 #if GEOS_DEBUG
-        cerr << " p1q1p2 && q1p2q2" << endl;
-#endif
-        intPt[0] = q1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q1z = interpolateZ(q1, p1, p2);
-        if(!std::isnan(q1z)) {
-            ztot += q1z;
-            hits++;
-        }
-        if(!std::isnan(q1.z)) {
-            ztot += q1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = p2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p2z = interpolateZ(p2, q1, q2);
-        if(!std::isnan(p2z)) {
-            ztot += p2z;
-            hits++;
-        }
-        if(!std::isnan(p2.z)) {
-            ztot += p2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " q1inP && p2inQ" << std::endl;
 #endif
+        // if pts are equal Z is chosen arbitrarily
+        intPt[0] = zGetOrInterpolateCopy(q1, p1, p2);
+        intPt[1] = zGetOrInterpolateCopy(p2, q1, q2);
+
 #if GEOS_DEBUG
-        cerr << " intPt[0]: " << intPt[0].toString() << endl;
-        cerr << " intPt[1]: " << intPt[1].toString() << endl;
+        std::cerr << " intPt[0]: " << intPt[0].toString() << std::endl;
+        std::cerr << " intPt[1]: " << intPt[1].toString() << std::endl;
 #endif
-        return (q1 == p2) && !p1q2p2 && !q1p1q2 ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
+        return (q1 == p2) && !q2inP && !p1inQ ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
     }
-    if(p1q2p2 && q1p1q2) {
+    if(q2inP && p1inQ) {
 #if GEOS_DEBUG
-        cerr << " p1q2p2 && q1p1q2" << endl;
-#endif
-        intPt[0] = q2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q2z = interpolateZ(q2, p1, p2);
-        if(!std::isnan(q2z)) {
-            ztot += q2z;
-            hits++;
-        }
-        if(!std::isnan(q2.z)) {
-            ztot += q2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = p1;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p1z = interpolateZ(p1, q1, q2);
-        if(!std::isnan(p1z)) {
-            ztot += p1z;
-            hits++;
-        }
-        if(!std::isnan(p1.z)) {
-            ztot += p1.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " q2inP && p1inQ" << std::endl;
 #endif
+        // if pts are equal Z is chosen arbitrarily
+        intPt[0] = zGetOrInterpolateCopy(q2, p1, p2);
+        intPt[1] = zGetOrInterpolateCopy(p1, q1, q2);
 #if GEOS_DEBUG
-        cerr << " intPt[0]: " << intPt[0].toString() << endl;
-        cerr << " intPt[1]: " << intPt[1].toString() << endl;
+        std::cerr << " intPt[0]: " << intPt[0].toString() << std::endl;
+        std::cerr << " intPt[1]: " << intPt[1].toString() << std::endl;
 #endif
-        return (q2 == p1) && !p1q1p2 && !q1p2q2 ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
+        return (q2 == p1) && !q1inP && !p2inQ ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
     }
-    if(p1q2p2 && q1p2q2) {
+    if(q2inP && p2inQ) {
 #if GEOS_DEBUG
-        cerr << " p1q2p2 && q1p2q2" << endl;
-#endif
-        intPt[0] = q2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        q2z = interpolateZ(q2, p1, p2);
-        if(!std::isnan(q2z)) {
-            ztot += q2z;
-            hits++;
-        }
-        if(!std::isnan(q2.z)) {
-            ztot += q2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[0].z = ztot / hits;
-        }
-#endif
-        intPt[1] = p2;
-#if COMPUTE_Z
-        ztot = 0;
-        hits = 0;
-        p2z = interpolateZ(p2, q1, q2);
-        if(!std::isnan(p2z)) {
-            ztot += p2z;
-            hits++;
-        }
-        if(!std::isnan(p2.z)) {
-            ztot += p2.z;
-            hits++;
-        }
-        if(hits) {
-            intPt[1].z = ztot / hits;
-        }
+        std::cerr << " q2inP && p2inQ" << std::endl;
 #endif
+        // if pts are equal Z is chosen arbitrarily
+        intPt[0] = zGetOrInterpolateCopy(q2, p1, p2);
+        intPt[1] = zGetOrInterpolateCopy(p2, q1, q2);
 #if GEOS_DEBUG
-        cerr << " intPt[0]: " << intPt[0].toString() << endl;
-        cerr << " intPt[1]: " << intPt[1].toString() << endl;
+        std::cerr << " intPt[0]: " << intPt[0].toString() << std::endl;
+        std::cerr << " intPt[1]: " << intPt[1].toString() << std::endl;
 #endif
-        return (q2 == p2) && !p1q1p2 && !q1p1q2 ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
+        return (q2 == p2) && !q1inP && !p1inQ ? POINT_INTERSECTION : COLLINEAR_INTERSECTION;
     }
     return NO_INTERSECTION;
 }
@@ -866,8 +552,8 @@ LineIntersector::intersection(const Coordinate& p1, const Coordinate& p2,
         //intPt = CentralEndpointIntersector::getIntersection(p1, p2, q1, q2);
         intPtOut = nearestEndpoint(p1, p2, q1, q2);
 #if GEOS_DEBUG
-        cerr << "Intersection outside segment envelopes, snapped to "
-             << intPt.toString() << endl;
+        std::cerr << "Intersection outside segment envelopes, snapped to "
+             << intPtOut.toString() << std::endl;
 #endif
     }
 
@@ -875,48 +561,116 @@ LineIntersector::intersection(const Coordinate& p1, const Coordinate& p2,
         precisionModel->makePrecise(intPtOut);
     }
 
-
-#if COMPUTE_Z
-    double ztot = 0;
-    double zvals = 0;
-    double zp = interpolateZ(intPtOut, p1, p2);
-    double zq = interpolateZ(intPtOut, q1, q2);
-    if(!std::isnan(zp)) {
-        ztot += zp;
-        zvals++;
-    }
-    if(!std::isnan(zq)) {
-        ztot += zq;
-        zvals++;
-    }
-    if(zvals) {
-        intPtOut.z = ztot / zvals;
-    }
-#endif // COMPUTE_Z
     return intPtOut;
 }
 
-/*private*/
-bool
-LineIntersector::isInSegmentEnvelopes(const Coordinate& pt) const
+/* private static */
+double
+LineIntersector::zInterpolate(const Coordinate& p, const Coordinate& p1, const Coordinate& p2)
 {
-    Envelope env0(*inputLines[0][0], *inputLines[0][1]);
-    Envelope env1(*inputLines[1][0], *inputLines[1][1]);
-    return env0.contains(pt) && env1.contains(pt);
-}
-
-/*private*/
-Coordinate
-LineIntersector::intersectionSafe(const Coordinate& p1, const Coordinate& p2,
-                                  const Coordinate& q1, const Coordinate& q2) const
-{
-    Coordinate ptInt = Intersection::intersection(p1, p2, q1, q2);
-    if (ptInt.isNull()) {
-        ptInt = nearestEndpoint(p1, p2, q1, q2);
+#if GEOS_DEBUG
+    std::cerr << " zInterpolate " << p << ", " << p1 << ", " << p2 << std::endl;
+#endif
+    double p1z = p1.z;
+    double p2z = p2.z;
+    if (std::isnan(p1z)) {
+      return p2z; // may be NaN
     }
-    return ptInt;
+    if (std::isnan(p2z)) {
+      return p1z; // may be NaN
+    }
+    if (p.equals2D(p1)) {
+      return p1z; // not NaN
+    }
+    if (p.equals2D(p2)) {
+      return p2z; // not NaN
+    }
+    double dz = p2z - p1z;
+    if (dz == 0.0) {
+      return p1z;
+    }
+#if GEOS_DEBUG
+    std::cerr << " Interpolating Z from distance of p along p1-p2" << std::endl;
+#endif
+    // interpolate Z from distance of p along p1-p2
+    double dx = (p2.x - p1.x);
+    double dy = (p2.y - p1.y);
+    // seg has non-zero length since p1 < p < p2
+    double seglen = (dx * dx + dy * dy);
+    double xoff = (p.x - p1.x);
+    double yoff = (p.y - p1.y);
+    double plen = (xoff * xoff + yoff * yoff);
+    double frac = sqrt(plen / seglen);
+    double zoff = dz * frac;
+    double zInterpolated = p1z + zoff;
+#if GEOS_DEBUG
+    std::cerr << " interpolated Z: " << zInterpolated << std::endl;
+#endif
+    return zInterpolated;
 }
 
 
+
+double
+LineIntersector::zInterpolate(const Coordinate& p,
+                              const Coordinate& p1,
+                              const Coordinate& p2,
+                              const Coordinate& q1,
+                              const Coordinate& q2)
+{
+#if GEOS_DEBUG
+    std::cerr << " zInterpolate(5 coords) called" << std::endl;
+#endif // GEOS_DEBUG
+    double zp = zInterpolate(p, p1, p2);
+#if GEOS_DEBUG
+    std::cerr << " zp: " << zp << std::endl;
+#endif // GEOS_DEBUG
+    double zq = zInterpolate(p, q1, q2);
+#if GEOS_DEBUG
+    std::cerr << " zq: " << zq << std::endl;
+#endif // GEOS_DEBUG
+    if (std::isnan(zp)) {
+      return zq; // may be NaN
+    }
+    if (std::isnan(zq)) {
+      return zp; // may be NaN
+    }
+#if GEOS_DEBUG
+    std::cerr << " averaging Z" << std::endl;
+#endif // GEOS_DEBUG
+    // both Zs have values, so average them
+    return (zp + zq) / 2.0;
+}
+
+
+/* private static */
+Coordinate
+LineIntersector::nearestEndpoint(const Coordinate& p1, const Coordinate& p2,
+                const Coordinate& q1, const Coordinate& q2)
+{
+    const Coordinate* nearestPt = &p1;
+    double minDist = Distance::pointToSegment(p1, q1, q2);
+
+    double dist = Distance::pointToSegment(p2, q1, q2);
+    if(dist < minDist) {
+        minDist = dist;
+        nearestPt = &p2;
+    }
+    dist = Distance::pointToSegment(q1, p1, p2);
+    if(dist < minDist) {
+        minDist = dist;
+        nearestPt = &q1;
+    }
+    dist = Distance::pointToSegment(q2, p1, p2);
+    if(dist < minDist) {
+        nearestPt = &q2;
+    }
+    return *nearestPt;
+}
+
 } // namespace geos.algorithm
 } // namespace geos
+
+#ifndef GEOS_INLINE
+# include "geos/algorithm/LineIntersector.inl"
+#endif
